%Plots the attractor of the hidden neural activation of the multiple model.
%This is done by running the simulator with the multiple model and saving the
%neural activation as we go. The plot shows the neural activation level for
%all the hidden nodes in successive pairs, i.e. if there are four nodes,
%plots will be shown for nodes 1-2, 2-3 and 3-4. In addition, a filtered
%attractor plot is shown, including neural activation only when the lambda
%value is above a given threshold. This will show how the attractors differ
%when they are actually contributing to the motor output, and when they are
%not (i.e. producing bad predictions).
%
%handles = plot_attractors(mm, mm_error_log, run_simulator, filename_prefix, epoch)
%
%Inputs: 
%
%mm: The multiple model structure.
%
%mm_error_log: The error log for the entire run.
%
%run_simulator: A function handle that will run the simulator.
%
%filename_prefix: A prefix that will be used to store the plot as a PDF
%file. If it is the empty matrix [], no files are created.
%
%epoch: The epoch we are analyzing. If not specified, defaults to the last
%plot of the error log. It is present, because there might be situations
%where simulations have been run and the neural networks have been stored
%as the training proceeded. 
%
%Outputs:
%
%handles: The handles to the figures.

function handles = plot_attractors(mm, mm_error_log, run_simulator, filename_prefix, epoch)

if nargin < 5
    epoch = length(mm_error_log);
end

filtered = true;

%This is the threshold for inclusion of the lambda activation in the
%attractor plot. If the lambda value of a module is above the threshold, it
%will be plotted in the attractor plots.
threshold = 0.1;

[ states, mms ] = run_simulator(mm, mm(1).desired_state, mm(1).context);

%The handles matrix. Initialized to the empty matrix, so we can concatenate
%them.
handles = [];

%The number of columns in the output will depend on if the filtering is to
%be shown or not.
plot_cols = 2;
if filtered
    plot_cols = 4;
end


%For each of the modules, plot the neural activation of the hidden layer of
%the forward and inverse models. 
for i=1:length(mm)
    handles = [ handles figure('Name', [ 'Attractors of module ' num2str(i) ]) ];
    %The number of plots depends on the 
    %size of the hidden layer. It is the same for the forward and
    %inverse module.
    num_plots = mm(1).inverse.size(2) - 1;
    
    %Collect all the activation levels at the hidden layer of the forward and inverse module.
    hidden_activation_forward = zeros(mm(1).forward.size(2), length(mms));
    hidden_activation_inverse = zeros(mm(1).inverse.size(2), length(mms));
    for j=1:length(mms)
        hidden_activation_forward(:,j) = mms{j}(i).forward.L{2};
        hidden_activation_inverse(:,j) = mms{j}(i).inverse.L{2};
    end

    
    %Do the filtering
    
    if filtered
        hidden_activation_forward_filtered = [];
        hidden_activation_inverse_filtered = [];

        %For this module: look at all its lambda activations, at the
        %timesteps where it is above the threshold, add the neural
        %activation to the filtered matrix.
        for j=1:length(mm_error_log(epoch).err(i).lambda)
            
            if mm_error_log(epoch).err(i).lambda(j) > threshold
                hidden_activation_forward_filtered = [hidden_activation_forward_filtered hidden_activation_forward(:,j) ];
                hidden_activation_inverse_filtered = [ hidden_activation_inverse_filtered hidden_activation_inverse(:,j) ];
                
                %disp(sprintf('module %i, activation level %3.2f at timestep %i included', i, mm_error_log(epoch).err(i).lambda(j), j))
            end
            
        end
    end

    %Exclude the initial transient trajectory, since the memory at t=0 is
    %blank.
    hidden_activation_forward = hidden_activation_forward(:,2:end);
    hidden_activation_inverse = hidden_activation_inverse(:,2:end);
    
    %If the first lambda activation was above the threshold, it was
    %included in the filtered version. However, we want to exclude the
    %first transient.
    if mm_error_log(epoch).err(i).lambda(1) > threshold
        hidden_activation_forward_filtered = hidden_activation_forward_filtered(:, 2:end);
        hidden_activation_inverse_filtered = hidden_activation_inverse_filtered(:, 2:end);
    end
       
    %Plot counter.
    plot_counter = 1;
    %Plot the attractors.
    for j=1:num_plots

        %Plot the forward model attractors.

        subplot(num_plots, plot_cols, plot_counter);

        plot(hidden_activation_forward(j,:), hidden_activation_forward(j+1,:), 'b');
        
        axis square
        
        if j == 1
            title('Forward model attractors');
        end

        xlim([0 1 ]);
        ylim([0 1 ]);
        
        xlabel(['Hidden node ' num2str(j) ]);
        ylabel(['Hidden node ' num2str(j+1) ]);
        
        plot_counter = plot_counter + 1;

        
        
        
        
        
        
        
        if filtered
        
            %Plot the FILTERED forward model attractors.
            
            subplot(num_plots, plot_cols, plot_counter);
            
            plot(hidden_activation_forward_filtered(j,:), hidden_activation_forward_filtered(j+1,:), 'b');
            
            axis square
            
            if j == 1
                title('Filtered forward model attractors');
            end
            
            xlim([0 1 ]);
            ylim([0 1 ]);
        
            xlabel(['Hidden node ' num2str(j) ]);
            ylabel(['Hidden node ' num2str(j+1) ]);
            
            plot_counter = plot_counter + 1;
            
        end
        
        
        
        
        
        %Plot the inverse model attractors.
        
        subplot(num_plots, plot_cols, plot_counter);
        
        plot(hidden_activation_inverse(j,:), hidden_activation_inverse(j+1,:), 'r');
        
        axis square

        if j == 1
            title('Inverse model attractors');
        end
                
        xlabel(['Hidden node ' num2str(j) ]);
        ylabel(['Hidden node ' num2str(j+1) ]);
        
        xlim([0 1 ]);
        ylim([0 1 ]);
        
        plot_counter = plot_counter + 1;

        
        
        if filtered

            %Plot the FILTERED inverse model attractors.
            
            subplot(num_plots, plot_cols, plot_counter);
        
            plot(hidden_activation_inverse_filtered(j,:), hidden_activation_inverse_filtered(j+1,:), 'r');
            
            axis square
            
            if j == 1
                title('Filtered inverse model attractors');
            end
            
            xlabel(['Hidden node ' num2str(j) ]);
            ylabel(['Hidden node ' num2str(j+1) ]);
            
            xlim([0 1 ]);
            ylim([0 1 ]);
            
            plot_counter = plot_counter + 1;
        end
        
    end
    
    %Position the plot
    set(handles(i), 'Position', [ 1 1 plot_cols*170 num_plots*170 ], 'PaperPositionMode', 'auto', 'PaperType', 'A4');
    

end       

if ~isempty(filename_prefix)
    for i=1:length(handles)
        saveas(handles(i), sprintf('%s_attractor_module_%i.pdf', filename_prefix, i));
    end
end
