%Activates the multiple models architecture. The actual state (x), desired
%state (x_desired) and contextual information (y) is given to all the 
%corresponding modules, and activation flows like this: 
%
%x + x_desired -> inverse model -> motor command.
%
%motor command + x -> forward model -> prediction of next state.
%
%rp * likelihood(forward model prediction error) -> responsibility signal (after normalization)
%
%responsibility signal * motor output command of each module.
%
%In addition, after the computation fo the motor commands, the winning
%value is increased proportionally, since the responsibility signal (i.e.
%lambda) is used to train the neural network. This allows for soft-max
%computation of the motor commands, but more error signal sent to the
%winning module. 
%
%Some consideration should be given to the timing indices: when the
%likelihood is to be calculated, the difference between next state and the
%current predicted state is passed to the likelihood function. Since we
%cannot see into the future, I interpret this as comparing the prediction
%at the *last* timestep with current state.
%
%[mm, motor_commands, lambda, forw_err, inv_err, motor_error_ratio] = hybrid_mosaic_activate(mm, x, x_desired, y)
%
%Inputs:
%
%mm: The multiple model structure.
%
%x: The current state.
%
%x_desired: The desired state (i.e. goal/target).
%
%y: Context information.
%
%Outputs:
% 
%mm: The multiple model structure, holding all the activation of its neural
%networks.
%
%motor_commands: The total motor commands computed by using the
%responsibility signals of each module.
%
%lambda: The responsibility signals of each of the modules.
%
%forw_err: The prediction error of the forward module (i.e. how well it
%predicted at the *last* timestep compared the current state).
%
%inv_err: The feedback error motor command.
%
%motor_error_ratio: The ratio of the motor error feedback command to the
%total motor error command, i.e. |ufb|/(|ufb| + |motor commands|)

function [mm, motor_commands, lambda, forw_err, inv_err, motor_error_ratio] = hybrid_mosaic_activate(mm, x, x_desired, y)

forw_err = zeros(mm(1).forward.size(end), length(mm));

for i=1:length(mm)
    %Activating the neural networks. By using the neural networks as
    %storage themselves, I will not have to declare new areas of memory for
    %holding variables all the time.
    
    mm(i).inverse = rnn_activate(mm(i).inverse, [ x ; x_desired ]);
    mm(i).forward = rnn_activate(mm(i).forward, [ x ; mm(i).inverse.L{3} ]);
    mm(i).rp = rnn_activate(mm(i).rp, y);
    
    %Comparing the actual state with what was predicted last time, which
    %will determine how well the forward model was able to predict.
    forw_err(:,i) = x - mm(i).old_forward;

    %It is the absolute value that gives the likelihood, but this is
    %done in the likelihood function. 
    lambda(i) = mm(i).rp.L{3}*likelihood( sum( forw_err(:,i)), mm(1).sigma);
end

%If all the lambda values are zero, make them 1 instead so the scaling will
%work. Or else NaN-values will appear.
if sum(lambda) == 0
    lambda = lambda + 1;
end

%Normalization.

lambda = lambda/sum(lambda);

%Also called u_fb, the feedback error. Matches how well the old desired
%state compares to the actual state we are in now.
inv_err = feedback_error(mm(1).old_x_desired, x, mm(1).K);
%disp( [ '\tinv_err: ' num2str(inv_err', ' %f')])

%Computing the final motor commands.
motor_commands = zeros( mm(1).inverse.size(end), 1);

%The inverse output is in the range [0,1], and is therefore scaled to
%[-1,1], allowing for movement in both directions. There is a gain
%multiplied to the output as well, allowing stronger motor commands to be
%issued. The output gain can now be different from the feedback error gain.
for i=1:length(mm)
    motor_commands = motor_commands + scale_velocities(mm(i).inverse.L{3}) .* lambda(i) .* mm(1).output_gain;
end

%Calculating the motor error ratio. Note that it is calculated based on
%absolute values of the motor command and feedback error motor command,
%since if they have different signs the ratio will be wrong. I.e. if motor
%command = -1 and ufb = 2, then 2/(-1 + 2) = 2, which is wrong. The correct
%solution is 2/(1 + 2) = 2/3, which is correct. 
motor_error_ratio = sum(abs(inv_err))/(sum(abs(inv_err)) + sum(abs(motor_commands)));

%Adding the feedback error.
motor_commands = motor_commands + inv_err;

%Storing the old forward computation. Note that this is not strictly
%necessary, since I could access the L{3}layer directly. However, I have
%chosen to have an old_forward variable for readability of the code. 
for i=1:length(mm)
    mm(i).old_forward = mm(i).forward.L{3};
end
mm(1).old_x_desired = x_desired;

%Print the output of the activation.
if mm(1).verbose
    disp( [ 'lambda: ' num2str(lambda, '%f ') ' executing: ' num2str(motor_commands', '%f ')  ])
end