%Trains the multiple model. This in turn trains each neural network. The
%main idea is that execution function will return the state *after* the motor
%commands have been applied to it, so that when calling execute
%MatLab will wait for the simulator to respond with the changes done in the
%environment, as a result of the applied motor commands. 
%
%[mm, err] = hybrid_mosaic_train(mm, x_desired, y)
%
%Inputs:
%
%mm: The multiple model structure.
%
%x_desired: The desired state.
%
%y: The contextual information.
%
%Outputs:
%
%mm: The multiple model structure, trained on one pass through all the
%desired states. 
%
%err: The error of the forward and inverse models, the lambda values and
%the ratio of feedback error to the motor commands. Ideally, the feedback
%error should converge to 0, which would imply that the forward and inverse
%model work perfectly.
%
%actual_state: This is the actual state that the system went through when
%training the model. This can then be compared with the desired state of
%the system, to see how the system performed at a certain epoch. 

function [mm, err, actual_state] = hybrid_mosaic_train(mm, x_desired, y)

%To keep it simple, every simulation should start at the same position. x
%equals state.
%
%Matlab automatically makes three structs with all the fields, so the
%execute function must be specified at an index. Calling the execute
%function returns the init state of Elton, which is later stored in the
%old_forward and old_x_desired slots. 
x = mm(1).execute([]);

%To initialize the context matrices of the model. 
mm = mm_init_context(mm, x);

%Logging the foward error for each module, the inverse error (i.e. the ufb
%command) and also the lambda value for each module. The lambda value must
%also be logged. All this is done at each timestep, so it will be possible
%to scrutinize the relationships between the forward error, the inverse
%error and the lambda values. 
%
%Note that it is not necessary to log the inverse error for each module, it
%is just done so for convenience. 
for i=1:length(mm)
    err(i).forward = zeros(1, length(x_desired));
    err(i).inverse = zeros(1, length(x_desired));
    err(i).lambda = zeros(1, length(x_desired));
    err(i).ratio = zeros(1, length(x_desired));
    err(i).rp = zeros(1, length(x_desired));
end

%Create space for the actual state of the system.
actual_state = zeros(size(x_desired));

for i=1:length(x_desired)
    %Compute the motor commands and the errors of the respective modules.
    [mm, motor_commands, lambda, forw_err, inv_err, motor_error_ratio] = hybrid_mosaic_activate(mm, x, x_desired(:,i), y(:,i));

    %Logging.
    for j=1:length(mm)
        err(j).forward(i) = sum(abs(forw_err(:,j))); %The absolute forward error.
        err(j).inverse(i) = sum(abs(inv_err));%Note how this is the same for every multiple model. This is only done for convenience.
        err(j).lambda(i) = lambda(j); %Log the lambda values for this module.
        err(j).ratio(i) = motor_error_ratio; %The same as above - each module gets stored the ratio of the ufb/motor command at each timestep. Just for simplicity.
        err(j).rp(i) = mm(j).rp.L{3}; %Log the responsibility predictor output for this module.
    end
    
    %To aid the separation: the rp with highest confidence will now be given an
    %increased confidence. This is stated in Haruno, Wolpert MOSAIC Model for
    %sensorimotor control, p. 2206 "with those models with large softmax values
    %receiving proportionally more of their error signal than modules with
    %small softmax". Without stating how, of course. Notice how the increase in
    %proportion to the winning module is done *after* the computation of the
    %motor command; i.e. it is true to the article saying that the module
    %receives more of its error signal, and that the separation should not
    %happen before calculating the motor commands. This way, the separation of
    %forward modules can be soft, whilst the training is hard - which will
    %further separate the module. This must happen after the logging of
    %the lambda values. Of course.

    [biggest, indices]= max(lambda);
    lambda(indices(1)) = mm(1).error_proportion*lambda(indices(1));
    lambda = lambda/sum(lambda);
    
    %Propagate the error, with the biased lambda values. 
    %
    %NOTE: THIS DIFFERS FROM THE PSEUDOCODE - in the pseudocode, the
    %training of the networks is done AFTER sending the motor commands to
    %the simulator. However, the order is of no importance at all, and it
    %is written here before the sending of the motor commands just for
    %aesthetic reasons (i.e. I want to be done with all the multiple model
    %programming before finally sending the motor output to the simulator.
    for j=1:length(mm)
        mm(j).inverse = rnn_train(mm(j).inverse, inv_err, lambda(j));
        mm(j).forward = rnn_train(mm(j).forward, forw_err(:,j), lambda(j));
        mm(j).rp = rnn_train(mm(j).rp, lambda(j));
    end
    
    %Execute the motor commands. The next time, the state returned from
    %breve will be x_t+1
    x = mm(1).execute(motor_commands);
    %Log the state of the system after the motor commands have been
    %applied.
    actual_state(:,i) = x;
end